# UpdateList.py

from gettext import gettext as _
import logging
import os
import json
import yaml
import shutil
import apt_pkg
from gi.repository import Gio
from .OriginFilter import UpdateListFilterCache

OUTPUT_CONFIG_PATH = '/var/lib/kylin-system-updater/json/'
INPUT_CONFIG_PATH = '/usr/share/kylin-update-desktop-config/config/'
IMPORTANT_LIST_PATH = '/var/lib/kylin-software-properties/template/important.list'

class LocalUpgradeDataList:
    """
    Represent the (potentially partial) results of an unattended-upgrades
    run
    """
    def __init__(self,                  
                upgrade_groups_pkgs={},                 
                upgrade_groups=[],    
                single_pkgs=[],
                adjust_pkgs=[],
                ):
        #可升级的组列表
        self.upgrade_groups = upgrade_groups
        #组列表中包含的包
        self.upgrade_groups_pkgs = upgrade_groups_pkgs
        #推送的可升级的单包
        self.single_pkgs = single_pkgs
        #调整版本列表 源过滤
        self.adjust_pkgs = adjust_pkgs
        #加版本号的升级包
        self.versoin_pkgs = {'single_upgrade':{}, 'groups_upgrade':{}}

class UpdateList():

    def __init__(self,parent):
        self.parent = parent

        #所有的组升级安装列表
        self.local_upgrade_data = LocalUpgradeDataList({},[],[],[])

        if 'XDG_CURRENT_DESKTOP' in os.environ:
            self.current_desktop = os.environ.get('XDG_CURRENT_DESKTOP')
        else:
            self.current_desktop = ''

        if 'XDG_DATA_DIRS' in os.environ and os.environ['XDG_DATA_DIRS']:
            data_dirs = os.environ['XDG_DATA_DIRS']
        else:
            data_dirs = '/usr/local/share/:/usr/share/'
            
        self.application_dirs = [os.path.join(base, 'applications')
                                for base in data_dirs.split(':')]

        #是否关闭源过滤
        # if (parent.options.close_filter == True or parent.configs.getWithDefault("SystemStatus", "isclosefilter", False) == True) and self.parent.is_disc_source == True:

        if self.parent.options.close_filter == False and self.parent.is_disc_source == False:
            #开启原过滤
            self.is_close_filter = False
            self.fu = UpdateListFilterCache(self.parent)
        else:
            logging.info("Close to Allowed origin fiter...")
            self.is_close_filter = True
            

    #清空上次输出的分组JSON文件
    def _empty_output_dir(self):
        try:
            #清空 升级列表
            self.parent.configs.setValue("AutoUpgrade","upgradelist",'',True)

            if not os.path.exists(OUTPUT_CONFIG_PATH):
                os.makedirs(OUTPUT_CONFIG_PATH)
                logging.info('making the ConfigPath(%s) is complete...',OUTPUT_CONFIG_PATH)
            else:
                shutil.rmtree(OUTPUT_CONFIG_PATH)
                os.makedirs(OUTPUT_CONFIG_PATH)
                logging.info('Emptying the ConfigPath(%s) is complete...',OUTPUT_CONFIG_PATH)
        except Exception as e:
                logging.error(str(e))

    #读取推送列表，判断分组和单包推送，再进行源过滤
    def _read_important_list(self,cache,pkgs_upgrade):
        header = ''
        desc = ''
        pkg_important_list = []
        pkg_list = []
        group_important_list = []

        # 获取importantlist 本次更新推送
        try:
            with open(IMPORTANT_LIST_PATH, 'r') as f:
                data = f.read()
            important_list = data.split()
            logging.info("The Server Push List: %a",important_list)

            for pkg_name in important_list:
                #检查是否在cache 没有在cache中属于组
                if pkg_name in cache:
                    pkg_obj = cache[pkg_name]
                    #在可升级的列表当中  此步骤为了排除已安装不需要升级的
                    if pkg_obj.is_installed:
                        if pkg_obj in pkgs_upgrade:
                            pkgs_upgrade.remove(pkg_obj)
                            pkg_list.append(pkg_obj)
                    else:
                        pkg_list.append(pkg_obj)
                else:
                    group_important_list.append(pkg_name)
            
            if pkg_list != []:
                install_list,upgrade_list = self._make_fiter_origin(pkg_list,True)
                pkg_important_list = install_list + upgrade_list

            logging.info("Push Single Packages: %a, Push Groups:%a",pkg_important_list,group_important_list)
            return True,group_important_list,pkg_important_list,header,desc
        except Exception as e:
            header = _("read important list failed")
            desc = str(e)
            logging.error(header + desc)
            return False,group_important_list,pkg_important_list,header,desc

    def _make_pkg_info_json(self,cache,pkgs_list):
        total_download_size = 0
        total_installed_size = 0
        pkgs_info_json = {}

        for pkg_name in pkgs_list:
            try:
                pkg = cache[pkg_name]
                #当前版本
                cur_version = getattr(pkg.installed, "version", '')
                new_version = getattr(pkg.candidate, "version", '')

                #获取下载大小
                download_size = getattr(pkg.candidate, "size", 0)
                installed_size = getattr(pkg.candidate, "installed_size", 0)

                total_download_size = total_download_size + download_size
                total_installed_size = total_installed_size + installed_size

                pkgs_info_json.update({pkg_name:{"cur_version":cur_version,"new_version":new_version,\
                                                "download_size":str(download_size),"install_size":str(installed_size)}})

            except Exception as e:
                logging.error("this package(%s) not in list and error mes:%s",pkg_name,e)

        pkgs_info_json.update({"total_download_size":str(total_download_size)})
        pkgs_info_json.update({"total_install_size":str(total_installed_size)})
        return pkgs_info_json

    #检查包是否在cache中 返回新得列表没 有安装的话才添加到列表
    def _check_pkg_in_cache(self,cache,pkgs_list):
        new_pkgs_list = []
        for pkg_name in pkgs_list:
            #检查是否在cache 以及 是否安装检查 
            if pkg_name in cache and not cache[pkg_name].is_installed:
                new_pkgs_list.append(pkg_name)
            else:
                pass
                # logging.info("this package(%s) not in list ",pkg_name)
        return new_pkgs_list

    def _make_group_output_json(self,data,data_yaml,upgrade_pkgs_json,install_pkgs_json):
        groups_base_info = {}
        output_json = {}
        
        #FIXME: 确定输出文件的文件名 以及放置位置
        output_config_name = OUTPUT_CONFIG_PATH + data['package'] + '.json'

        #4、添加一些基础信息
        groups_base_info.update({"package":data['package']})
        groups_base_info.update({"new_version":data['version']})
        groups_base_info.update({"name":data['name']})
        groups_base_info.update({"description":data['description']})
        groups_base_info.update({"icon":data['icon']})
        
        #添加读yaml文件
        groups_base_info.update({"changelog":data_yaml['changelog']})

        #5、添加升级的内容
        output_json.update(groups_base_info)
        output_json.update({"upgrade_list":upgrade_pkgs_json})
        output_json.update({"install_list":install_pkgs_json})
        # output_json.update({"hold_list":hold_pkgs_list})
        # output_json.update({"remove_list":remove_pkgs_list})

        #6 产生JSON文件
        with open(output_config_name, 'w', encoding='utf-8') as f:
            json.dump(output_json, f, ensure_ascii=False, indent=4)                   
        logging.info("Generate Jsonfile(%s) to complete... ",output_config_name)

    #进行源过滤，is_adjust 是否调整cache中的候选版本，单包推送会调整保持控制面板显示正确的版本
    def _make_fiter_origin(self,_all_pkg_obj,is_adjust):
        new_install_list = []
        new_upgrade_list = []

        adjust_pkgs = []
        all_pkg_obj = []

        #是否进行源过滤的选项
        if self.is_close_filter == False:
            try:
                all_pkg_obj,adjust_pkgs = self.fu.check_in_allowed_origin(_all_pkg_obj,is_adjust)
            except Exception as e:
                all_pkg_obj = _all_pkg_obj
                adjust_pkgs = []
                logging.error("Check Allowed origin is occur error:" + str(e))
                logging.error("Close to Allowed origin fiter...")
        else:
            all_pkg_obj = _all_pkg_obj
            adjust_pkgs = []

        self.local_upgrade_data.adjust_pkgs += adjust_pkgs

        for pkg_obj in all_pkg_obj:
            if pkg_obj.is_installed:
                new_upgrade_list.append(pkg_obj.name)
            else:
                new_install_list.append(pkg_obj.name)
        return new_install_list,new_upgrade_list

    #从本地中获取本次升级需要升级的包 部分升级和全部升级使用 全盘升级不适用
    def _make_pkgs_list(self,cache,groups_pkgs,groups_list,pkg_list):
        pkgs_install = []
        pkgs_upgrade = []
        
        try:
            #单包的升级方式
            for pkg in pkg_list:
                if cache[pkg].is_installed:
                    pkgs_upgrade.append(pkg)
                else:
                    pkgs_install.append(pkg) 

            #遍历升级组列表
            for group_name in groups_list:
                pkgs_install += groups_pkgs.get(group_name,[]).get('pkgs_install',[])
                pkgs_upgrade += groups_pkgs.get(group_name,[]).get('pkgs_upgrade',[]) 
            
            return pkgs_install,pkgs_upgrade
        except Exception as e:
            logging.error(str(e))
            return [],[]      
    
    #输出白名单的配置
    def _make_autoupgrade_config(self,cache,upgrade_data,_adjust_pkgs):
        try:
            pkgs_install,pkgs_upgrade = self._make_pkgs_list(cache,upgrade_data.upgrade_groups_pkgs,upgrade_data.upgrade_groups,upgrade_data.single_pkgs)
            
            split_adjust_pkgs = [i.split("=")[0] for i in _adjust_pkgs]

            pkgs_upgrade_str = ''
            pkgs_list = pkgs_upgrade + pkgs_install
            if pkgs_list:
                for pkg_str in pkgs_list:
                    
                    #记录源过滤后调整的版本
                    if pkg_str in split_adjust_pkgs:
                        pkg_str = _adjust_pkgs[split_adjust_pkgs.index(pkg_str)]
                    else:
                        ver = getattr(cache[pkg_str].candidate, "version", '')
                        pkg_str = pkg_str + "=" + ver
                    pkgs_upgrade_str += (pkg_str + ',')

                pkgs_upgrade_str=pkgs_upgrade_str[:-1]

            self.parent.configs.setValue("AutoUpgrade","upgradelist",pkgs_upgrade_str,True)
            logging.info("Generate AutoUpgrade Configfile to Complete and Upgradable Number:%d...",len(pkgs_list))
        except Exception as e:
            logging.error(e)

    def _make_groups_upgrade(self,cache,group_list, pkgs_upgrade = []): 
        try:
            if os.path.isdir(INPUT_CONFIG_PATH) == False:
                logging.warning("configPath(%s) is not exists...",INPUT_CONFIG_PATH)
                return

            files = os.listdir(INPUT_CONFIG_PATH) #获得文件夹中所有文件的名称列表

            for ifile in files:
                #判是否是目录以及是否以JSON结尾
                if ifile.endswith('.json'):
                    #读取组JSON文件
                    with open(INPUT_CONFIG_PATH+ifile,'r') as f:
                        try:
                            data = json.load(f)
                        except json.JSONDecodeError as exc:
                            logging.error(exc)
                            continue

                    group_name = data['package']
                    #读取组的yaml 文件的changelog的信息
                    with open(INPUT_CONFIG_PATH + group_name + ".yaml", "r") as stream:
                        try:
                            data_yaml = yaml.safe_load(stream)
                        except yaml.YAMLError as exc:
                            logging.error(exc)
                            continue

                    #过滤没有推送的配置文件
                    if not group_name in group_list:
                        continue
                    upgrade_pkgs_list =  data['upgrade_list']
                    # hold_pkgs_list = data['hold_list']
                    #这个安装升级列表中包含当前系统的cache中没有的包 需要过滤
                    # remove_pkgs_list = data['remove_list']

                    #检查包是否在cache中 以及是否已经安装 没有安装的话才添加到列表
                    new_install_list = self._check_pkg_in_cache(cache,data['install_list'])
                    #进行交集 升级列表
                    new_upgrade_list = list(set(pkgs_upgrade) & set(upgrade_pkgs_list))

                    #进行源过滤 
                    new_install_list,new_upgrade_list = self._make_fiter_origin([cache[pkg] for pkg in new_install_list + new_upgrade_list],False)
                    
                    #单包的优先级最高 从组中剔除此包
                    for pkg in self.local_upgrade_data.single_pkgs:
                        if pkg in new_install_list:
                            new_install_list.remove(pkg)

                    #判断当前是否可升级或者新装的包
                    if len(new_install_list) == 0 and len(new_upgrade_list) == 0:
                        continue
                    #在总升级列表中移除这些包
                    for pkg in new_upgrade_list:
                        pkgs_upgrade.remove(pkg)
                    #3、生成升级的包列表JSON
                    upgrade_pkgs_json = self._make_pkg_info_json(cache,new_upgrade_list)
                    #2、生成安装的软件列表
                    install_pkgs_json = self._make_pkg_info_json(cache,new_install_list)
                    #输出JSON配置文件
                    self._make_group_output_json(data,data_yaml,upgrade_pkgs_json,install_pkgs_json)

                    #保存分组版本号,好像没有
                    self.local_upgrade_data.versoin_pkgs['groups_upgrade'].update({group_name:''})
                    
                    #添加到字典维护的升级列表
                    self.local_upgrade_data.upgrade_groups.append(group_name)
                    self.local_upgrade_data.upgrade_groups_pkgs.update({group_name:{"pkgs_upgrade":new_upgrade_list,"pkgs_install":new_install_list}})
                    logging.info("Group(%s) upgrade:%d install:%d",group_name,len(new_upgrade_list),len(new_install_list))
                else:
                    pass
        except Exception as e:
            logging.warning("Generate Jsonfile to failed... ")
            logging.error(e)

    def _rate_application_for_package(self, application, pkg):
        score = 0
        desktop_file = os.path.basename(application.get_filename())
        application_id = os.path.splitext(desktop_file)[0]

        if application.should_show():
            score += 1

            if application_id == pkg.name:
                score += 5

        return score

    def _file_is_application(self, file_path):
        # WARNING: This is called often if there's a lot of updates. A poor
        # performing call here has a huge impact on the overall performance!
        if not file_path.endswith(".desktop"):
            # First the obvious case: If the path doesn't end in a .desktop
            # extension, this isn't a desktop file.
            return False

        file_path = os.path.abspath(file_path)
        for app_dir in self.application_dirs:
            if file_path.startswith(app_dir):
                return True
        return False

    def _get_application_for_package(self, pkg):
        desktop_files = []
        rated_applications = []

        for installed_file in pkg.installed_files:
            if self._file_is_application(installed_file):
                desktop_files.append(installed_file)

        for desktop_file in desktop_files:
            try:
                application = Gio.DesktopAppInfo.new_from_filename(
                    desktop_file)
                application.set_desktop_env(self.current_desktop)
            except Exception as e:
                logging.warning("Error loading .desktop file %s: %s" %
                                (desktop_file, e))
                continue
            score = self._rate_application_for_package(application, pkg)
            if score > 0:
                rated_applications.append((score, application))

        rated_applications.sort(key=lambda app: app[0], reverse=True)
        if len(rated_applications) > 0:
            return rated_applications[0][1]
        else:
            return None

    def _make_single_upgrade(self,cache,pkg_list):
        try:
            for pkg in pkg_list:
                zh_name = ''
                base_info = {}
                output_json = {}
                output_config_name = OUTPUT_CONFIG_PATH + pkg + '.json'
                
                pkg_cache = cache[pkg]

                #获取包的软件名称，从主题中读取，只有可升级，是软件的包才可以读取
                if pkg_cache.is_installed:
                    app = self._get_application_for_package(pkg_cache)
                    if app is not None:
                        zh_name = app.get_display_name()
                
                pkgs_json = self._make_pkg_info_json(cache,[pkg])
                en_name = getattr(pkg_cache.candidate, "summary", '')
                description_str = getattr(pkg_cache.candidate, "description", '')

                #4、添加一些基础信息
                base_info.update({"package":pkg})

                base_info.update({"cur_version":getattr(pkg_cache.installed, "version", '')})
                base_info.update({"new_version":getattr(pkg_cache.candidate, "version", '')})
                base_info.update({"name":{"zh_CN":zh_name,"en_US":en_name}})
                base_info.update({"description":{"zh_CN":description_str,"en_US":description_str}})
                base_info.update({"icon":''})

                #5、添加升级的内容
                output_json.update(base_info)
                if pkg_cache.is_installed:
                    output_json.update({"upgrade_list":pkgs_json})
                    output_json.update({"install_list":{}})
                else:
                    output_json.update({"upgrade_list":{}})
                    output_json.update({"install_list":pkgs_json})
                
                #产生JSON文件
                with open(output_config_name, 'w', encoding='utf-8') as f:
                    json.dump(output_json, f, ensure_ascii=False, indent=4)                
                logging.info("Generate Jsonfile(%s) to complete... ",output_config_name)

                #6、保存单包版本号
                self.local_upgrade_data.versoin_pkgs['single_upgrade'].update({pkg_cache.name:getattr(pkg_cache.installed, "version", '')})
        except Exception as e:
            logging.warning("Generate Jsonfile to failed... ")
            logging.error(str(e))

    def update(self,cache):
        pkgs_upgrade = []
        header = ''
        desc = ''

        #查找所有可升级的包
        for pkg in cache:
            try:
                if pkg.is_upgradable and pkg.is_installed:
                    if getattr(pkg.candidate, "origins", None) is None:
                        logging.error("WARNING: upgradable but no candidate.origins?!?: ",
                            pkg.name)
                        continue
                    pkgs_upgrade.append(pkg)
            except Exception as e:
                logging.error(e)
        
        logging.info("System all upgradeable packages:upgrade:%d ",len(pkgs_upgrade))

        success,group_important_list,self.local_upgrade_data.single_pkgs,header,desc = self._read_important_list(cache,pkgs_upgrade)

        #清空输出的目录
        self._empty_output_dir()

        #important_list 为空时此次不需要升级
        if success == True and not group_important_list and not self.local_upgrade_data.single_pkgs:
            header = _("The software on this computer is up to date.")
            desc = ""
            return success,header,desc
        elif success == False:
            return success,header,desc

        #产生单包的JSON
        self._make_single_upgrade(cache,self.local_upgrade_data.single_pkgs)

        #分组的包的JSON
        self._make_groups_upgrade(cache,group_important_list,[pkg.name for pkg in pkgs_upgrade])

        self._make_autoupgrade_config(cache,self.local_upgrade_data,self.local_upgrade_data.adjust_pkgs)
        #是否存在可升级的组
        if self.local_upgrade_data.upgrade_groups or self.local_upgrade_data.single_pkgs:
            return True,header,desc
        else:
            header = _("The software on this computer is up to date.")
            desc = ''
            return True,header,desc

