#!/usr/bin/python3
import os
import dbus
import dbus.service
import logging

import apt_pkg
import subprocess#snap
from gettext import gettext as _
from .backend import InstallBackend
from .Core.loop import mainloop
import SystemUpdater.Core.enums as enums
from .Core.utils import humanize_size,PolicyKit_Authority,get_proc_from_dbus_name
from SystemUpdater.Core.utils import unLockedEnableShutdown,deb_verify
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
import locale
from importlib import reload
from .Core.MyCache import MyCache

UPDATER_DBUS_INTERFACE = 'com.kylin.systemupgrade.interface'
UPDATER_DBUS_PATH = '/com/kylin/systemupgrade'
UPDATER_DBUS_SERVICE = 'com.kylin.systemupgrade'
RUN_UNATTENDED_UPGRADE = '/var/run/unattended-upgrades.pid'

#颜色设置
COLORLOG_SUFFIX = "\033[0m"

# Define some foreground colors
BLACK = 30
RED = 31
GREEN = 32
YELLOW = 33
BLUE = 34
MAGENTA = 35
CYAN = 36
WHITE = 37

#字体颜色
FRONT_COLOR_SEQ = "\033[1;%dm"
#背景颜色
BACK_COLOR_SEQ = "\033[%d;1m"

COLORLOG_PREFIX     = FRONT_COLOR_SEQ % GREEN                                                                 
COLORMETHOR_PREFIX  = FRONT_COLOR_SEQ % CYAN

UU_UPGRADE_MODE_TIMING = 0
UU_UPGRADE_MODE_BEFORE_SHUTDOWN = 1

#dbus 建立
class UpdateManagerDbusController(dbus.service.Object):
    """ this is a helper to provide the UpdateManagerIFace """

    RETURN_SUCCESS_CODE = 0
    RETURN_SUCCESS_DESC = ""

    RETURN_UNKNOWN_CODE = -1
    RETURN_UNKNOWN_DESC = ""

    RETURN_BUSY_STATE = 1
    RETURN_BUSY_DESC = "In the process of updating or Upgrading..."

    def __init__(self, parent, bus_name,
                object_path=UPDATER_DBUS_PATH):
        dbus.service.Object.__init__(self, bus_name, object_path)
        self.parent = parent
        self.bus = dbus.SystemBus()

        self.transaction = None

    #更新important.list的本次升级的列表
    def on_update_important_list(self):
        header = None
        desc = None

        try:
            # logging.info("Start to update Source Template...")
            obj = self.bus.get_object('com.kylin.software.properties', '/com/kylin/software/properties')
            interface = dbus.Interface(obj, dbus_interface='com.kylin.software.properties.interface')
            retval = interface.updateSourceTemplate(timeout=20)
        except Exception as e:
            header = _("Unable to access the source management server")
            desc = str(e)
            logging.error("update sourceTemplate Failed and Error mes:%s"%str(e))
            return False,header,desc
        
        if retval == False:
            header = _("Unable to access the source management server")
            desc = _("Connection exception with push server")
            logging.warning(header + desc)
            return retval,header,desc
        else:
            # logging.info("Update sourceTemplate successed...")
            return retval,header,desc

    def is_reboot_required(self):
        """If a reboot is required to get all changes into effect."""
        return os.path.exists(os.path.join(apt_pkg.config.find_dir("Dir"),
                                        "var/run/reboot-required"))

    def is_logout_required(self):
        """If a logout is required to get all changes into effect."""
        return os.path.exists(os.path.join(apt_pkg.config.find_dir("Dir"),
                                        "var/run/logout-required"))

    #重启aptdeamon后台服务
    def make_aptdeamon_restart(self):
        header = ''
        desc = ''

        try:
            obj = self.bus.get_object('org.debian.apt', '/org/debian/apt')
            interface = dbus.Interface(obj, dbus_interface='org.debian.apt')
            retval = interface.Quit()
        except Exception as e:
            header = _("aptdeamon restart occur Exception")
            logging.error(str(e))
            return False,header,desc

    #设置aptdeamon的环境变量
    def set_aptdeamon_environ(self,key,value):
        try:
            logging.info("Set aptdeaom environment variables %s = %s...",key,value)
            obj = self.bus.get_object('org.debian.apt', '/org/debian/apt')
            interface = dbus.Interface(obj, dbus_interface='org.debian.apt')
            retval = interface.SetEnviron(key,value)
            return retval
        except Exception as e:
            logging.error(str(e))
            return False

    def check_connectivity(self):
        header = ''
        desc = ''

        try:
            obj = self.bus.get_object("org.freedesktop.NetworkManager","/org/freedesktop/NetworkManager")
            interface = dbus.Interface(obj, "org.freedesktop.NetworkManager")
            retval = interface.CheckConnectivity(timeout=0.5)
            #1 表示没有网卡可以使用
            if retval == 1:
                header = _("Please check your network connection and retry.")
                return False,header,desc
            else:
                return True,header,desc

        except Exception as e:
            # logging.error(str(e))
            logging.info("Check Connectivity timeout and Determination Network is working...")
            #调用超时 也认为是网络正常
            return True,header,desc
    
    #当检查失败时 再切换到ping再进行一次检查
    def _check_network_ping(self):
        header = ''
        desc = ''
        slist = apt_pkg.SourceList()
        slist.read_main_list()

        #多个源 也只循环一次
        for uri in [lis.uri for lis in slist.list]:
            #只要是使用光盘源 就不进行检查了 默认检查网络成功 进行之后的动作
            if "file://" in uri:
                logging.info("Current exites of CD-ROM source in sources.list and  return network ok...")
                return True,header,desc

            if "localhost:9977" in uri:
                #deb http://localhost:9977/172.17.126.249:8098/deb/kylin/ 需要去除localhost:9977 在apt-p2p的模式下
                network_uri = uri.split("//")[1].split("/")[1]
            else:
                #只拿去 网站主地址 例如http://ppa.launchpad.dev/luoxueyi/v101/kylin-desktop 只截取ppa.launchpad.dev 使用此来检查网络
                network_uri = uri.split("//")[1].split("/")[0]

            #去除这个问题172.17.126.249:8098 带端口号的 
            if ":" in network_uri:
                network_uri = network_uri.split(":")[0]
            
            logging.info("Check network connectivity to use ping %s...",network_uri)
            #ping 一次，总超时时间为5S
            response = os.system("ping -c 1 -w 5 "+network_uri+" 2>&1 >/dev/null")
            if response == 0:
                return True,header,desc
            else:
                header = _("Please check your network connection and retry.")
                return False,header,desc
        #当源列表为空时 从此反回，这种认为这是一种正常的状况 进行之后的动作 
        return True,header,desc

    def _check_prohibit_user(self, sender_name):
        prohibit_list = ["dbus-send","gdbus"]
        if sender_name in prohibit_list:
            raise dbus.exceptions.DBusException("ERROR: You are not allowed to perform this action.")

    @dbus.service.method(UPDATER_DBUS_INTERFACE,
                        in_signature="", out_signature="",
                        sender_keyword="caller_name")
    def Quit(self, caller_name):
        """Request a shutdown of the daemon."""
        #如果在下载就请求 取消
        self.CancelDownload()
        try:
            #尝试接触 关闭枷锁
            unLockedEnableShutdown()
        except:
            logging.error("File lock release failure")
        logging.info("Quitting was requested")
        logging.debug("Quitting main loop...")
        mainloop.quit()
        logging.debug("Exit")

    #Remove all downloaded files.
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def Clean(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('Clean In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' Clean ...')
                self.parent.start_clean()
                return True
        except Exception:
            return False

    #获取后端现在的状态
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='s',out_signature='i')
    def GetBackendStatus(self,user_lang):
        try:
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetBackendStatus and user_lang = %s ...',str(user_lang))
            #当传入为空时 直接返回
            if str(user_lang) == '':
                logging.info("The incoming language is null...")
            else:
                (lang, encoding) = locale._parse_localename(str(user_lang))
                
                #合法性的校验
                if lang == None or lang == '':
                    logging.info("The incoming language is Illegal...")
                    return self.parent.is_working

                #处于更新和升级中的话 不进行更新
                if lang != os.environ["LANGUAGE"]:
                    logging.warning("The system language is switched %s to %s",os.environ["LANGUAGE"],lang)
                    os.environ["LANGUAGE"] = lang

            return self.parent.is_working
        except Exception as e:
            logging.error(str(e))
            return False

    #apt install -f
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def FixBrokenDepends(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('FixBrokenDepends In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' FixBrokenDepends ...')
                self.parent.start_fix_broken()
                return True
        except Exception:
            return False

    #dpkg configure -a 
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def FixIncompleteInstall(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('FixIncompleteInstall In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' FixIncompleteInstall ...')
                self.parent.start_fix_incomplete()
                return True
        except Exception:
            return False

    #更新的dbus
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='b')
    def UpdateDetect(self):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('UpdateDetect In the process of Updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UpdateDetect ...')
                self.parent.start_update()
                return True
        except Exception:
            return False
    
    #部分升级
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='bas',out_signature='bs')
    def DistUpgradePartial(self,_is_install,_partial_upgrade_list):
        try:
            is_install = bool(_is_install)
            if not self.parent.update_list:
                logging.warning('Perform \"UpdateDetect\" first')
                return False,'Perform \"UpdateDetect\" first'
            #处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('DistUpgradePartial In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                partial_upgrade_list = [str(i) for i in _partial_upgrade_list]
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' DistUpgradePartial and is_install:%r, partial_upgrade_list:%s. ',is_install,",".join(partial_upgrade_list))
                local_upgrade_groups = self.parent.update_list.local_upgrade_data.upgrade_groups
                local_single_pkgs = self.parent.update_list.local_upgrade_data.single_pkgs

                new_upgrade_list = list(set(partial_upgrade_list) & set(local_upgrade_groups + local_single_pkgs))

                if new_upgrade_list:
                    self.parent.start_install(InstallBackend.MODE_UPGRADE_PARTIAL,is_install,partial_upgrade_list)
                    return True,'dbus upgrading'
                else:
                    logging.warning('input upgrade list(%s) not in local upgrade_list(%s)',partial_upgrade_list,local_upgrade_groups+local_single_pkgs)
                    return False,'upgrade_list is empty'
        except Exception as e:
            return False,e

    #全部升级
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='bs')
    def DistUpgradeAll(self,_is_install):
        try:
            is_install = bool(_is_install)
            if not self.parent.update_list:
                logging.warning('Perform \"UpdateDetect\" first')
                return False,'Perform \"UpdateDetect\" first'
            #处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('DistUpgradeAll In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' DistUpgradeAll and is_install:%r...',is_install)
                self.parent.start_install(InstallBackend.MODE_UPGRADE_ALL,is_install)
                return True,'success'
        except Exception as e:
            return False,str(e)
            
    # 全盘升级 也就是 apt dist-upgrade 升级的方式
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='b',out_signature='bs')
    def DistUpgradeSystem(self,_is_install):
        try:
            is_install = bool(_is_install)
            #处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('DistUpgradeSystem In the process of updating or Upgrading...')
                return False,'In the process of updating or Upgrading...'
            else:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' DistUpgradeSystem and is_install:%r...',is_install)
                self.parent.start_install(InstallBackend.MODE_UPGRADE_SYSTEM,is_install)
                return True,'success'
        except Exception as e:
            logging.error(False, str(e))
            return (False, str(e))

    #卸载包
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='asss',out_signature='b',sender_keyword='sender')
    def PurgePackages(self,_purge_list,cur_user,user_lang = '',sender=None):
        try:
            #当传入为空时 直接返回
            if str(user_lang) == '':
                logging.info("The incoming language is null...")
            else:
                (input_lang, encoding) = locale._parse_localename(str(user_lang))
                (local_lang, encoding) = locale._parse_localename(str(os.environ["LANG"]))
                (local_language, encoding) = locale._parse_localename(str(os.environ["LANGUAGE"]))
                
                #合法性的校验
                if input_lang == None or input_lang == '':
                    logging.info("The incoming language is Illegal...")
                    return self.RETURN_UNKNOWN_CODE,"The incoming language is Illegal..."

                #处于更新和升级中的话 不进行更新
                if input_lang != local_language or input_lang != local_lang:
                    logging.warning("LANG: switched %s to %s",os.environ["LANG"],user_lang)
                    logging.warning("LANGUAGE: switched %s to %s",os.environ["LANGUAGE"],user_lang)
                    os.environ["LANG"] = user_lang
                    os.environ["LANGUAGE"] = user_lang
                    self.parent.aptd_lang_switch = True
                    reload(enums)
            purge_list = [str(pkg) for pkg in _purge_list]
            sender_name = get_proc_from_dbus_name(sender)
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' DistPurgePackages Sender:%s and purge list is:%s...',sender_name, purge_list)
            PolicyWhiteList = ["kylin-uninstaller","kylin-software-center"]
            if sender_name not in PolicyWhiteList:
                (status, details) = PolicyKit_Authority(enums.get_source_name_from_enum(sender_name)+_(" requires authentication to uninstall software packages."), sender)
                if not status:
                    self.PurgePackagesFinished(False,details,'')
                    return False

            #目前只有360使用这个环境变量 当其他包也使用时 可以将这个权限放开
            # if "360epp" in purge_list:
            if True:
                #需要对aptdeamon加这两个环境变量 才可以提示弹窗
                self.set_aptdeamon_environ("XAUTHORITY","/home/"+str(cur_user)+"/.Xauthority")
                self.set_aptdeamon_environ("DISPLAY",":0")

            # 处于更新和升级中的话 不进行升级
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                self.PurgePackagesFinished(False,_("Other tasks are being updated and upgraded, please uninstall them later."),'')
                logging.warning('PurgePackages In the process of updating or Upgrading...')
                return False
            else:
                # logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' DistPurgePackages and purge list is:%s...',purge_list)
                self.parent.start_purge_pkgs(purge_list)
                return True
        except Exception as e:
            return False,str(e)

    # 取消transaction 
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='b')
    def CancelDownload(self):
        status = False
        try:
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' CancelDownload...')
            
            #空时直接返回
            if self.transaction == None:
                logging.info("transaction is none")
                return False
            
            if self.transaction.cancellable == True:
                self.transaction.cancel()
                status = True
                logging.info("dbus-mothod cancel task Success")
            else:
                status = False
                logging.info("cancel task Failed")
                
        except Exception as e:
            logging.error("cancel task Failed" + str(e))
            return status
        return status


    # 安装本地deb包
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='ssbbs',out_signature='b',sender_keyword='sender')
    def InstallDebFile(self,source = "unKnown", path = "", _check_local_dep = False, _auto_satisfy = False, user_lang = '', sender=None):
        try:
            #当传入为空时 直接返回
            if str(user_lang) == '':
                logging.info("The incoming language is null...")
            else:
                (input_lang, encoding) = locale._parse_localename(str(user_lang))
                (local_lang, encoding) = locale._parse_localename(str(os.environ["LANG"]))
                (local_language, encoding) = locale._parse_localename(str(os.environ["LANGUAGE"]))
                
                #合法性的校验
                if input_lang == None or input_lang == '':
                    logging.info("The incoming language is Illegal...")
                    return self.RETURN_UNKNOWN_CODE,"The incoming language is Illegal..."

                #处于更新和升级中的话 不进行更新
                if input_lang != local_language or input_lang != local_lang:
                    logging.warning("LANG: switched %s to %s",os.environ["LANG"],user_lang)
                    logging.warning("LANGUAGE: switched %s to %s",os.environ["LANGUAGE"],user_lang)
                    os.environ["LANG"] = user_lang
                    os.environ["LANGUAGE"] = user_lang
                    self.parent.aptd_lang_switch = True
                    reload(enums)
            sender_name = get_proc_from_dbus_name(sender)
            self._check_prohibit_user(sender_name)

            check_local_dep = bool(_check_local_dep)
            auto_satisfy = bool(_auto_satisfy)
            deb_path = str(path)
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' InstallDebFile and check_local_dep:%r, auto_satisfy:%r, current_lang:%s , InstallDebFile sender: %s .',\
                                        check_local_dep,auto_satisfy,user_lang,sender_name)
            logging.info("Will install: %s.",path)
            self.parent.start_deb_install(deb_path, _check_local_dep, _auto_satisfy, source, sender)
            return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC
        except Exception as e:
            logging.error(str(e))
            return self.RETURN_UNKNOWN_CODE,str(e)
    
    # commitpackages
    @dbus.service.method(UPDATER_DBUS_INTERFACE,in_signature='as',out_signature='is',sender_keyword='sender')
    def InstallPackages(self, pkg_list = [], sender=None):
        try:
            sender_name = get_proc_from_dbus_name(sender)
            self._check_prohibit_user(sender_name)
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' InstallPackages : [%s],  InstallPackages sender: %s .', ", ".join(list(pkg_list)), sender_name)
            self.parent.start_back_upgrade(pkg_list, "kylin-background-upgrade")
            return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC
        except Exception as e:
            logging.error(str(e))
            return self.RETURN_UNKNOWN_CODE,str(e)

    #更新的dbus
    @dbus.service.method(UPDATER_DBUS_INTERFACE,out_signature='is',sender_keyword='sender')
    def UpdateCache(self,sender=None):
        try:
            #处于更新和升级中的话 不进行更新
            if self.parent.is_working != InstallBackend.ACTION_DEFUALT_STATUS:
                logging.warning('UpdateDetect In the process of Updating or Upgrading...')
                return self.RETURN_BUSY_STATE,self.RETURN_BUSY_DESC
            else:
                sender_name = get_proc_from_dbus_name(sender)
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UpdateCache sender:%s...',sender_name)
                self._check_prohibit_user(sender_name)
                self.parent.start_update_backend(InstallBackend.MODE_UPDATE_CACHE)
                return self.RETURN_SUCCESS_CODE,self.RETURN_SUCCESS_DESC
        except Exception as e:
            logging.error(str(e))
            return self.RETURN_UNKNOWN_CODE,str(e)

    # dbus接口：向数据库display表中插入数据
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='ss', out_signature='b', sender_keyword='sender')
    def InsertInstallState(self, item, value, sender=None):
        try:
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' InsertInstallState, options:%s，value:%s' % (item, value))
            self.parent.sqlite3_server.insert_into_display(item, value)
            logging.info("Database inserted successfully，options:%s，value:%s" % (item, value))
            return True
        except Exception as e:
            logging.error("Database insert failed，options:%s，value:%s" % (item, value))
            return False
    
    # # download certain package and its dependencies
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='asi', out_signature='b')
    def DownloadPackages(self, pkgs=[], try_times=1):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' DownloadPackages: %s',pkgs)
        if try_times > 0:
            pass
        else:
            try_times = 1
        times = try_times
        while times > 0:
            self.parent.start_download(pkgs)
            times -= 1
        return True

    # 是否允许关机前更新
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='ss', out_signature='bs')
    def UnattendedUpgradeValue(self, operation, value="false"):
        if operation.lower() != "get" and operation.lower() != "set":
            return False, 'Please input [\"set\", \"value\"] to set. \nor [\"get\"] to get whether updates are allowed before shutdown.'
        if operation == "set":
            try:
                logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UnattendedUpgradeValue is going to %s [allow_unattended_upgrades_shutdown] value to %s.'%(operation,value))
                self.parent.sqlite3_server.insert_into_display("allow_unattended_upgrades_shutdown", value.lower())
            except Exception as e:
                logging.error(str(e))
                return False,str(e)
        else:
            logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' UnattendedUpgradeValue is going to %s [allow_unattended_upgrades_shutdown] value.'%(operation))
            try:
                value = self.parent.sqlite3_server.select_from_display("allow_unattended_upgrades_shutdown")
                logging.info("[allow_unattended_upgrades_shutdown] value is %s."%(value))
            except Exception as e:
                logging.error(str(e))
                return False,str(e)
        return True,"success"

    # 设置自动更新时间
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='bs')
    def SetAutoUpgradePeriod(self, period):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutoUpgradePeriod will be set value %s.'%(period))
        try:
            self.parent.uuconfigs.setValue("autoUpgradePolicy", "updateDays", str(period), True)
        except Exception as e:
            logging.error(str(e))
        return True,"success"

    # 设置自动更新时间
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='bs')
    def SetAutoUpgradeRandomRange(self,randomRange):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetAutoUpgradeRandomRange will be set value %s.'%(randomRange))
        try:
            self.parent.uuconfigs.setValue("autoUpgradePolicy", "randomRange", str(randomRange), True)
        except Exception as e:
            logging.error(str(e))
        return True,"success"
    
    # # dbus接口：改变apt下载速度
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='sb', out_signature='b')
    def SetDownloadspeedMax(self, speed, set):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetDownloadspeedMax, speed:%s, set:%r.'%(speed, set))
        
        #来重启Aptdeamon
        self.parent.set_restart_aptdeamon = True
        if set:
            with open("/etc/apt/apt.conf.d/80apt-download", "w+") as f:
                try:
                    f.write("Acquire::http::Dl-Limit" + " \"" + "%s" % str(speed) + "\";\n")
                    f.write("Acquire::https::Dl-Limit" + " \"" + "%s" % str(speed) + "\";\n")
                    #更改数据库值
                    self.parent.sqlite3_server.insert_into_display("download_limit","true")
                    self.parent.sqlite3_server.insert_into_display("download_limit_value",str(speed))
                    #发送信号
                    self.DownloadSpeedChanged(str(speed))
                    return True
                except Exception as e:
                    logging.error(e)
                    return False
        else:
            if os.path.exists("/etc/apt/apt.conf.d/80apt-download"):
                os.remove("/etc/apt/apt.conf.d/80apt-download")
                self.parent.sqlite3_server.insert_into_display("download_limit","false")
                self.DownloadSpeedChanged("0")
                return True
            else:
                self.parent.sqlite3_server.insert_into_display("download_limit","false")
                self.DownloadSpeedChanged("0")
                return True

    # # dbus接口：获取apt下载速度
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='bs')
    def GetDownloadspeedLimitValue(self):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetDownloadspeedLimitValue...')
        try:
            download_limit = self.parent.sqlite3_server.select_from_display("download_limit")
            if download_limit == "true":
                download_limit_value = self.parent.sqlite3_server.select_from_display("download_limit_value")
                return True,str(download_limit_value)
            else:
                return False,str("0")
        except:
            return False, "0"
        
    ## dbus接口: 开启或关闭预下载功能
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='b', out_signature='b')
    def SetPreDownloadState(self, _state):
        state = bool(_state)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetPreDownloadState, state is %r ...',state)
        try:
            if state:
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "preDownload", "on", True)
            else :
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "preDownload", "off", True)
            self.ChangeUpgradePolicy()
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    ## dbus接口: 发送立即更新的信号
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='b')
    def AutoUpgradeAllNow(self):
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' AutoUpgradeAllNow ...')
        try:
            self.UpgradeAllNow()
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    ## dbus接口: 开启关闭自动更新功能
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='b', out_signature='b')
    def SetAutoUpgradeState(self, _state):
        state = bool(_state)
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetAutoUpgradeState, state is %r ...',state)
        try:
            if state:
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "autoUpgradeState", "on", True)
                self.parent.sqlite3_server.insert_into_display("autoupdate_allow", "true")
            else :
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "autoUpgradeState", "off", True)
                # self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadMode", "manual", True)
                # self.parent.uuconfigs.setValue("autoUpgradePolicy", "installMode", "manual", True)
                self.parent.sqlite3_server.insert_into_display("autoupdate_allow", "false")
            self.ChangeUpgradePolicy()
            
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    ## dbus接口: 设置自动更新策略
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='is', out_signature='b')
    def SetAutoUpgradeMode(self, mode, time):
        logging.info(COLORMETHOR_PREFIX+'method'+COLORLOG_SUFFIX+' SetAutoUpgradeMode, mode is %s, time is %s ...',mode, time)
        try:
            if mode == UU_UPGRADE_MODE_TIMING:
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadMode", "timing", True)
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadTime", str(time), True)
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "installMode", "timing", True)
            elif mode == UU_UPGRADE_MODE_BEFORE_SHUTDOWN:
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "downloadMode", "timing", True)
                self.parent.uuconfigs.setValue("autoUpgradePolicy", "installMode", "bshutdown", True)
            self.ChangeUpgradePolicy()
        except Exception as e:
            logging.error(str(e))
            return False
        return True
    
    # # dbus接口: 后端大数据采集
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='ss', out_signature='b')
    def DataBackendCollect(self, messageType, uploadMessage):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' DataBackendCollect, messageType is %s ...',messageType)
        try:
            self.parent.collector.UpdateMsg(messageType, uploadMessage)
        except AttributeError as e:
            logging.error(str(e))
        except Exception as e:
            logging.error(str(e))

    # 检查更新
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='as', out_signature='as', sender_keyword='sender')
    def CheckInstalledOrUpgrade(self, pkgs, sender=None):
        sender_name = get_proc_from_dbus_name(sender)
        try:
            # logging.info("CheckInstalledOrUpgrade: pkgs %s.", ", ".join(pkgs))
            pkglist = []
            cache = MyCache(None)
            self.parent.collector.cache = cache
            for i in pkgs:
                name = str(i).strip().replace("\n", "")
                if name in cache:
                    pkg = cache[name]
                    if pkg.is_installed == False or pkg.is_upgradable == True:
                        pkglist.append(pkg.name)
                        pkglist.append(str(pkg.candidate.size))
                        # self.parent.collector.make_background_version(pkg)
                        # self.parent.collector.background_upgradable.append(pkg.name)
        except:
            logging.error("CheckInstalledOrUpgrade: Failed to obtain package information: %s" % str(i), exc_info=True)
        # logging.info("CheckInstalledOrUpgrade: pkglist %s.", ", ".join(pkglist))
        return pkglist

    # # dbus接口：安装本地snap包
    # @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='b', sender_keyword='sender')
    # def InstallSnap(self, path, sender=None):
    #     logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+" Install Snap Packages(%s)...",path)
    #     try:
    #         # 提权
    #         sender_name = get_proc_from_dbus_name(sender)
    #         if deb_verify(path) != 0: #验签失败，提权
    #             (status,error_string) = PolicyKit_Authority(sender_name+_(" requires authentication to install software packages."),sender)
    #             if not status:
    #                 msg = {"statue": "Fail", "errorcode": str(0), "error_message": _("Authentication failure.")}
    #                 kwarg = {"error_message": str(msg)}
    #                 self.KumSnapSignal("snap_error", kwarg)
    #                 return False
    #         status= subprocess.getstatusoutput("snap install " +path+" --dangerous")
    #     except Exception as e:
    #         logging.error("Install Snap Packages occur error:%s",str(e))
    #         msg = {"statue": "Fail", "errorcode": str(0), "error_message": _("Command execution error")}
    #         kwarg = {"error_message": str(msg)}

    #         self.KumSnapSignal("snap_error", kwarg)
    #         return False

    #     #安装成功
    #     if status[0]==0:
    #         logging.info("Install Snap Packages(%s) is success",path)
    #         msg = {"statue": "Success", "errorcode": str(0), "error_message": ""}
    #         kwarg = {"action": str(msg)}
    #         self.KumSnapSignal("snap_finish", kwarg)
    #         return True
    #     else:
    #         #失败之后的处理
    #         logging.info("Install Snap Packages(%s) is Failed",path)
    #         if "supported architectures" in status[1]:
    #             logging.error("Supported architectures does not match")
    #             msg = {"statue": "Fail", "errorcode": str(0), "error_message": _("Unsupported architecture")}
    #         else:
    #             logging.error("Other Error...")
    #             msg = {"statue": "Fail", "errorcode": str(0), "error_message": _("Other Error")}
    #         kwarg = {"error_message": str(msg)}

    #         self.KumSnapSignal("snap_error", kwarg)
    #         return False

    # 监控是否需要重启的操作
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='s', out_signature='b')
    def CheckRebootRequired(self, model):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' CheckRebootRequired is be %s',str(model))
        try:
            #更新完成之后检查是否需要重启
            if self.is_reboot_required() == True:
                self.RebootLogoutRequired("reboot")
                logging.info("System need to Reboot...")
            elif self.is_logout_required() == True:
                self.RebootLogoutRequired("logout")
                logging.info("System need to Logout...")
            else:
                logging.info("System not need to Reboot or Logout...")
        except Exception as e:
            logging.error(str(e))
            return False
        return True

    # kill 进程
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='i', out_signature='b')
    def KillProcessSignal(self, pid):
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' KillProcessSignal is %d', pid)
        try:
            # 判断文件是否存在
            if (os.path.exists(RUN_UNATTENDED_UPGRADE)):
                os.kill(int(pid), 9)
                logging.info('%s has been killed', pid)
            else:
                logging.warning('%s is not exist.', RUN_UNATTENDED_UPGRADE)
        except Exception as e:
            logging.error(str(e))
            return False
        return True
    
    # 获取数据库值
    @dbus.service.method(UPDATER_DBUS_INTERFACE, in_signature='bss', out_signature='s')
    def GetSetDatabaseInfo(self, gs, table, field):
        Text = 'NULL'
        try:
            if gs: #get 
                if table == 'display':
                    Text = self.parent.sqlite3_server.select_from_display(str(field))
                    logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetDatabaseInfo Table:%s Field:%s Text:%s',table,field,Text)
            else: #set 
                if table == 'display' and "=" in field:
                    field, value = str(field).split("=")
                    logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' SetDatabaseInfo Table:%s Field:%s', table, field)
                    self.parent.sqlite3_server.insert_into_display(field, value)
                    return "success"
        except Exception as e:
            logging.error(str(e))
            return Text
        return Text
    
    # 获取系统版本号
    @dbus.service.method(UPDATER_DBUS_INTERFACE, out_signature='ss',sender_keyword='sender')
    def GetSystemUpdateVersion(self,sender=None):
        sender_name = get_proc_from_dbus_name(sender)
        logging.info(COLORMETHOR_PREFIX+'Method'+COLORLOG_SUFFIX+' GetSystemUpdateVersion sender: %s .', sender_name)
        os_version = ''
        update_version = ''
        try:
            self.parent.sqlite3_server.ucconfigs = UpgradeConfig(datadir = "/etc/kylin-version", name = "kylin-system-version.conf")
            if not os.path.exists("/etc/kylin-version/kylin-system-version.conf"): 
                logging.warning("System version file doesn't exist.")
                update_version,os_version = self.parent.sqlite3_server.get_default_version()
                return str(os_version),str(update_version)
            os_version = str(self.parent.sqlite3_server.ucconfigs.get("SYSTEM","os_version"))
            update_version = str(self.parent.sqlite3_server.ucconfigs.get("SYSTEM","update_version"))
        except Exception as e:
            logging.error(str(e))
            return str(e),''
        logging.info('Current os_version: %s, release_id: %s .', os_version, update_version)
        return os_version,update_version

    #更新进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='is')
    def UpdateDetectStatusChanged(self,progress,status):
        logging.info(COLORLOG_PREFIX+"Emitting"+COLORLOG_SUFFIX+" UpdateDetectStatusChanged progress = %d , status = %s",progress,status)
    
    #替代UpdateDetectFinished 分发更新完成信号，为了解决发送空列表时 控制面板出段错误
    def DistUpdateDetectFinished(self, success, upgrade_group,error_string='',error_desc=''):
        if success == False and upgrade_group == []:
            upgrade_group = ['']
        self.UpdateDetectFinished(success,upgrade_group,error_string,error_desc)
    #更新完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='basss')
    def UpdateDetectFinished(self, success, upgrade_group,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " UpdateDetectFinished success = %r , upgrade_group = %a, error_string = %s , error_desc = %s ",\
                    success,upgrade_group,error_string,error_desc)
    
    #升级的进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='asiss')
    def UpdateDloadAndInstStaChanged(self,groups_list,progress,status,current_details):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" UpdateDloadAndInstStaChanged upgrade groups_list = %s progress = %d , status = %s ,current_details = %s\033[0m",\
            groups_list,progress,status,current_details)

    #升级完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='basss')
    def UpdateInstallFinished(self, success, upgrade_group,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" UpdateInstallFinished success = %r , upgrade_group = %a, error_string = %s , error_desc = %s ",\
                    success,upgrade_group, error_string,error_desc)

    #发送下载包信息 fix bug 字节大小改成u 无符号32位
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='asiiuui')
    def UpdateDownloadInfo(self,upgrade_group,current_items, total_items, currenty_bytes, total_bytes, current_cps):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" UpdateDownloadInfo upgrade_group = %a,current_items = %d, total_items = %d, currenty_bytes = %s, total_bytes = %s, current_cps = %s/s",\
                    upgrade_group,\
                    current_items, total_items, \
                    humanize_size(currenty_bytes), humanize_size(total_bytes),\
                    humanize_size(current_cps))

    #查询解决依赖 信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bbasasasss')
    def UpdateDependResloveStatus(self, resolver_status, remove_status,remove_pkgs,pkg_raw_description,delete_desc,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" UpdateDependResloveStatus:resolver_status = %r , remove_status = %r , remove_pkgs = %a,pkg_raw_description = %s ,delete_desc = %s,error_string = %s , error_desc = %s ",\
            resolver_status,remove_status,remove_pkgs,pkg_raw_description,delete_desc,error_string,error_desc)

    #查询dist-upgrade解决依赖
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bbasasasss')
    def DistupgradeDependResloveStatus(self, resolver_status, remove_status,remove_pkgs,pkg_raw_description,delete_desc,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" DistupgradeDependResloveStatus:resolver_status = %r , remove_status = %r , remove_pkgs = %a,pkg_raw_description = %s ,delete_desc = %s ,error_string = %s , error_desc = %s ",\
            resolver_status,remove_status,remove_pkgs,pkg_raw_description,delete_desc,error_string,error_desc)

    # 信号是否可取消
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='b')
    def Cancelable(self, Cancelable):
        pass
        # logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" Cancelable: %r",Cancelable)
    
    # 插入数据库完成后发送
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='ss')
    def UpdateSqlitSingle(self, appname, date):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" UpdateSqlitSingle: [ %s ]: date: %s .",\
                    appname, date)

    #更新进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bbisss')
    def FixBrokenStatusChanged(self,finished,success,progress,status,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX+"Emitting"+COLORLOG_SUFFIX+" FixBrokenStatusChanged finished = %r , success = %r,progress = %d , status = %s,error_string = %s , error_desc = %s",\
            finished,success,progress,status,error_string,error_desc)

    #卸载完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bss')
    def PurgePackagesFinished(self, success,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " PurgePackagesFinished success = %r , error_string = %s , error_desc = %s ",\
                    success,error_string,error_desc)

    #卸载进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='iss')
    def PurgePkgStatusChanged(self,progress,status,current_details):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" PurgePkgStatusChanged progress = %d , status = %s ,current_details = %s",\
            progress,status,current_details)
    # dbus 信号：用于发送立即更新信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE)
    def UpgradeAllNow(self):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" UpgradeAllNow") 
    
    # dbus 信号：用于发送自动更新配置更改信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE)
    def ChangeUpgradePolicy(self):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" ChangeUpgradePolicy") 
    #安装deb包完成的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='bss')
    def InstalldebFinished(self, success,error_string='',error_desc=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " InstalldebFinished success = %r , error_string = %s , error_desc = %s ",\
                    success,error_string,error_desc)
        
    #安装进度信息 0~100 进度信息 101为非预期的信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='iss')
    def InstalldebStatusChanged(self,progress,status,current_details):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" InstalldebStatusChanged progress = %d , status = %s ,current_details = %s",\
            progress,status,current_details)

    # dbus 信号：用于发送本地安装snap包的结果
    @dbus.service.signal(UPDATER_DBUS_INTERFACE, signature='sa{sv}')
    def KumSnapSignal(self, type_msg, msg):
        logging.info(COLORLOG_PREFIX + "Emitting" + COLORLOG_SUFFIX +" KumSnapSignal type_msg = %s,msg = %s",type_msg,str(msg))  

    #重启和注销请求信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='s')
    def RebootLogoutRequired(self,required_status=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " RebootLogoutRequired required_status = %s",required_status)\
        
    #限速修改信号
    @dbus.service.signal(UPDATER_DBUS_INTERFACE,signature='s')
    def DownloadSpeedChanged(self,limit_value=''):
        logging.info(COLORLOG_PREFIX + "Emitting"+ COLORLOG_SUFFIX + " DownloadSpeedChanged limit_value = %s",limit_value)