#!/usr/bin/env python
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-

"""Integration of package managers into SystemUpdater"""

import logging
import os
from gettext import gettext as _

import apt
from SystemUpdater.Core.utils import (
    unLockedEnableShutdown,
    check_free_space,
    get_broken_details
)
from apt import Cache
from SystemUpdater.Core.UpdateList import LocalUpgradeDataList
import threading
from SystemUpdater.Core.utils import emulate_calcul_broken
import subprocess
import threading

class NowUpgradePara:
    """
    Represent the (potentially partial) results of an unattended-upgrades
    run
    """
    def __init__(self,                  
                upgrade_groups,                 
                upgrade_mode,    
                single_pkgs,
                ):
        #组列表中包含的包
        self.upgrade_groups = upgrade_groups

        #升级的模式
        self.upgrade_mode = upgrade_mode
        #推送的可升级的单包
        self.single_pkgs = single_pkgs

class InstallBackend():
    ACTION_DEFUALT_STATUS = -1
    ACTION_UPDATE = 0
    ACTION_INSTALL = 1
    ACTION_INSTALL_DEB = 2
    ACTION_CHECK_RESOLVER = 3
    ACTION_DOWNLOADONLY = 4
    ACTION_FIX_BROKEN = 5
    ACTION_REMOVE_PACKAGES = 6
    ACTION_FIX_INCOMPLETE = 7
    ACTION_CLEAN = 8
    ACTION_INSTALL_SHUTDOWN = 9
    ACTION_BACKGROUND_UPGRADE = 10
    ACTION_CHECK_BROKEN = 11

    MODE_DEFAULT_STATUS = -1
    
    #1、ACTION_INSTALL 安装的子类
    #部分升级
    MODE_UPGRADE_PARTIAL = 0
    #全部升级
    MODE_UPGRADE_ALL = 1
    #系统全盘升级
    MODE_UPGRADE_SYSTEM = 2
    #后端内部安装包使用 目前 更新配置包和升级本身使用
    MODE_UPGRADE_SINGLE = 3

    #2、更新的子类
    MODE_UPDATE_CACHE = 0
    MODE_UPDATE_ALL = 1

    def __init__(self, window_main, action,mode = MODE_DEFAULT_STATUS):
        self.window_main = window_main
        self.cache = window_main.cache
        self.action = action
        self.action_mode = mode

        #对所有的操作进行枷锁 当执行完毕后返回到_action_done 进行解锁，除了计算依赖模式 不进入_action_done 单独解锁
        self.window_main.is_working = action

        #更新的时候此对象还未生成
        if self.window_main.update_list !=  None:
            #本地维护的升级列表和需要升级的包
            self.upgrade_data = window_main.update_list.local_upgrade_data
        else:
            #生成一个空的列表 防止出现错误 在更新和安装单包的时候这个对象还没生成
            self.upgrade_data = LocalUpgradeDataList({},[],[],[])

        #要拿到升级列表必须调用_make_upgrade_list
        self.now_upgrade = NowUpgradePara([],self.MODE_DEFAULT_STATUS,[])

        self.trans_error_details = ''
        
        #当下载数量大于100个包时  就认为属于大版本升级 开启重试机制
        self.is_version_upgrade = False
        #True表示需要重试安装
        self.is_need_retry = False

    def start(self,_upgrade_mode = MODE_DEFAULT_STATUS,partial_upgrade_list = []):
        _success = True
        header = ''
        desc = ''

        #安装升级包 首先必须调用ACTION_CHECK_RESOLVER 计算依赖解决方便 标记cache 进行升级
        if self.action == self.ACTION_INSTALL:
            self.now_upgrade.upgrade_mode = _upgrade_mode

            if _upgrade_mode != self.MODE_UPGRADE_SINGLE:
                #拿到升级列表
                self.now_upgrade.upgrade_groups,self.now_upgrade.single_pkgs = self._make_groups_list(self.upgrade_data,_upgrade_mode,partial_upgrade_list)

            pkgs_install,pkgs_upgrade,pkgs_remove = self._get_mark_from_cache(self.cache,self.upgrade_data.adjust_pkgs,_upgrade_mode)

            logging.info("INSTALL install:%d , upgrade:%d remove:%d",len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
            
            #当下载数量大于200个包时  就认为属于大版本升级 开启重试机制
            if len(pkgs_install) + len(pkgs_upgrade) > 100:
                logging.info("Open a major version upgrade and Retry mechanism on...")
                self.is_version_upgrade = True
            
            #检查是否存在可升级的包
            if len(pkgs_install) == 0 and len(pkgs_upgrade) == 0 and len(pkgs_remove) == 0:
                #FIXME: 没有需要升级和安装的包  检查升级包是否不存在本次升级
                header = _("There is an exception in the update package.")
                desc = _("This update cannot detect the upgradeable package.")
                self._action_done(self.action,False,False,header,desc)
                return

            #检查磁盘的状态 
            _success,desc = check_free_space(self.cache)
            if _success == False:
                header = _("Disk space is insufficient, please clean the disk and then upgrade")
                #当单包安装检查内存不够的时候，报错到控制面板 单包安装属于更新才使用的所以直接赋值为更新 再进行之后的报错处理
                if self.now_upgrade.upgrade_mode == self.MODE_UPGRADE_SINGLE:
                    self.action = self.ACTION_UPDATE

                self._action_done(self.action,False,False,header,desc)
                return
                
            logging.info("Disk Check finished...")
            try:
                self.commit(pkgs_install, pkgs_upgrade, pkgs_remove)
            except Exception as e:
                logging.error(e)
        #计算依赖解决方案
        elif self.action == self.ACTION_CHECK_RESOLVER:
            #被删除的包列表
            delete_pkgs = []
            #被删除包的描述
            raw_description = []
            #那些包升级导致删除包 全盘升级不包含 
            delete_desc = []
            
            #检查依赖之前 判断是否需要重启aptdeamon 目前为了生效限速功能
            self.window_main.check_restart_aptdeamon()

            if _upgrade_mode != self.MODE_UPGRADE_SYSTEM:
                #获取要升级的组列表
                self.now_upgrade.upgrade_groups,self.now_upgrade.single_pkgs = self._make_groups_list(self.upgrade_data,_upgrade_mode,partial_upgrade_list)
                #获取要升级和安装的包列表
                pkgs_install,pkgs_upgrade = self._make_pkgs_list(self.cache,self.upgrade_data.upgrade_groups_pkgs,self.now_upgrade.upgrade_groups,self.now_upgrade.single_pkgs)
                #计算解决依赖关系
                _success,delete_pkgs,delete_desc,header,desc = self._make_problem_resolver(self.cache,pkgs_install,pkgs_upgrade)
                pkgs_install,pkgs_upgrade,pkgs_remove = self._get_mark_from_cache(self.cache,self.upgrade_data.adjust_pkgs,_upgrade_mode)
                
                if len(pkgs_remove) != len(delete_pkgs):
                    logging.warning("Simulation of the deletion package list:%s",str(delete_pkgs))
                    logging.warning("ProblemResolver of the deletion package list:%s",str(pkgs_remove))
                    delete_desc = []
                else:
                    pkgs_remove = delete_pkgs
            else:
                # 使用全盘升级 全盘使用dist-upgrade
                try:
                    if self.cache.get_changes():
                        self.cache.clear()
                    _success = self.cache._depcache.upgrade(True)
                    pkgs_install,pkgs_upgrade,pkgs_remove = self._get_mark_from_cache(self.cache,self.upgrade_data.adjust_pkgs,_upgrade_mode)
                except Exception as e:
                    _success = False
                    desc = str(e)
                    logging.error(desc)

            logging.info("RESOLVER install:%d , upgrade:%d remove:%d",len(pkgs_install),len(pkgs_upgrade),len(pkgs_remove))
            is_remove_pkgs = len(pkgs_remove) != 0

            #添加关于删除包的描述信息
            for pkg in pkgs_remove:
                pkg_obj = self.cache[pkg]
                # app = UpdateList.get_application_for_package(pkg_obj)
                # if app is not None:
                #     raw_description.append(app.get_display_name())
                # else:
                raw_description.append(getattr(pkg_obj.candidate, "summary", ''))
            #补充删除描述
            if pkgs_remove != [] and delete_desc == []:
                for pkg in pkgs_remove:
                    delete_desc.append('')
            
            #解锁后端状态
            self.window_main.is_working = self.ACTION_DEFUALT_STATUS

            if _upgrade_mode != self.MODE_UPGRADE_SYSTEM:
                self.window_main.dbusController.UpdateDependResloveStatus(_success,is_remove_pkgs,pkgs_remove,raw_description,delete_desc,header,desc)
            else:
                self.window_main.dbusController.DistupgradeDependResloveStatus(_success,is_remove_pkgs,pkgs_remove,raw_description,delete_desc,header,desc)
        #修复未满足的依赖关系 与apt-get -f install匹配
        elif self.action == self.ACTION_FIX_BROKEN:
            self.fix_broken()
        # 修复不完整的安装dpkg configure -a
        elif self.action == self.ACTION_FIX_INCOMPLETE:
            self.fix_incomplete()
        #卸载包
        elif self.action == self.ACTION_REMOVE_PACKAGES:
            self.purge_packages(partial_upgrade_list)
        #清空所有下载的文件
        elif self.action == self.ACTION_CLEAN:
            self.clean()
        #更新cache
        elif self.action == self.ACTION_UPDATE:
            #检查依赖之前 判断是否需要重启aptdeamon 目前为了生效限速功能
            self.window_main.check_restart_aptdeamon()
            self.update()

    def start_alone(self,partial_upgrade_list = [],_is_install = False):
        #单独下载功能
        if self.action == self.ACTION_DOWNLOADONLY: 
            try:
                self.download_deb(partial_upgrade_list)
            except Exception as e:
                logging.error(str(e))
        # 安装本地deb包的接口
        elif self.action == self.ACTION_INSTALL_DEB: 
            try:
                self.install_deb(install_path = partial_upgrade_list, install_force = _is_install)
            except Exception as e:
                logging.error(str(e))
        # 安装在线包的接口
        elif self.action == self.ACTION_BACKGROUND_UPGRADE:
            try:
                pkgs_install = [ str(pkg) for pkg in partial_upgrade_list]
                logging.info("Install deb package, open cache")
                cache = Cache()
                self.window_main.sqlite3_server.deb_metadata.update({"background_list":pkgs_install})
                for pkg_name in pkgs_install:
                    if pkg_name not in cache:
                        if "=" not in pkg_name or ("=" in pkg_name and pkg_name.split("=")[0] not in cache):
                        # 没找到包或格式不正确
                            self.window_main.dbusController.UpdateInstallFinished(False, pkgs_install, "'"+pkg_name+"' is not in cache", "")
                            return
                    else:
                        pkg = cache[pkg_name]
                        if pkg.installed:
                            self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].update({pkg.name:str(pkg.installed.version)})
                        else:
                            self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].update({pkg.name:str("")})
                self.commit_background(pkgs_install,[],[])
            except Exception as e:
                logging.error(str(e))
                self.window_main.dbusController.UpdateInstallFinished(False, pkgs_install, str(e), "")

    def update(self):
        """Run a update to refresh the package list"""
        raise NotImplementedError

    def clean(self):
        """Remove all downloaded files.t"""
        raise NotImplementedError

    def commit(self, pkgs_install, pkgs_upgrade, pkgs_remove,pkgs_purge = []):
        """Commit the cache changes """
        raise NotImplementedError

    def install_deb(self,install_path,install_force):
        """install_deb the cache changes """
        raise NotImplementedError

    def fix_broken(self):
        """fix_broken_depends"""
        raise NotImplementedError

    def fix_incomplete(self):
        """fix_incomplete_install"""
        raise NotImplementedError

    def purge_packages(self,pkgs_purge):
        """purge_packages"""
        raise NotImplementedError

    #从cache中拿到标记的列表
    def _get_mark_from_cache(self,cache,_adjust_pkgs,upgrade_mode):
        pkgs_install = []
        pkgs_upgrade = []
        pkgs_remove = []

        #全盘升级不做任何的调整
        if upgrade_mode == self.MODE_UPGRADE_SYSTEM:
            adjust_pkgs = []
        else:
            #获取调整包列表 去掉版本号
            adjust_pkgs = [i.split("=")[0] for i in _adjust_pkgs]

        for pkg in cache:
            try:
                if pkg.marked_install:
                    pkgname = pkg.name
                    if pkg.is_auto_installed:
                        pkgname += "#auto"
                    if pkg.name in adjust_pkgs:
                        pkgs_install.append(_adjust_pkgs[adjust_pkgs.index(pkg.name)])
                    else:
                        pkgs_install.append(pkgname)
                elif pkg.marked_upgrade:
                    if pkg.name in adjust_pkgs:
                        pkgs_upgrade.append(_adjust_pkgs[adjust_pkgs.index(pkg.name)])
                    else:
                        pkgs_upgrade.append(pkg.name)
                elif pkg.marked_delete:
                    pkgs_remove.append(pkg.name)
            except KeyError:
                pass
        return pkgs_install,pkgs_upgrade,pkgs_remove

    #获取当前升级的升级列表
    def _make_groups_list(self,upgrade_data,_upgrade_mode,partial_upgrade_list):
        groups_list = []
        pkg_list = []
        
        #部分升级
        if _upgrade_mode == self.MODE_UPGRADE_PARTIAL:
            for elem in partial_upgrade_list:
                #组升级方式
                if elem in upgrade_data.upgrade_groups:
                    groups_list.append(elem)
                #单包升级方式
                elif elem in upgrade_data.single_pkgs:
                    pkg_list.append(elem)
                else:
                    logging.warning("this package(%s) not in selected list",elem)
            
        #全部升级列表
        elif _upgrade_mode == self.MODE_UPGRADE_ALL:
            groups_list = upgrade_data.upgrade_groups
            pkg_list = upgrade_data.single_pkgs
        
        return groups_list,pkg_list

    #从本地中获取本次升级需要升级的包 部分升级和全部升级使用 全盘升级不适用
    def _make_pkgs_list(self,cache,groups_pkgs,groups_list,pkg_list):
        pkgs_install = []
        pkgs_upgrade = []
        
        try:
            #单包的升级方式
            for pkg in pkg_list:
                if cache[pkg].is_installed:
                    pkgs_upgrade.append(pkg)
                else:
                    pkgs_install.append(pkg) 

            #遍历升级组列表
            for group_name in groups_list:
                pkgs_install += groups_pkgs.get(group_name,[]).get('pkgs_install',[])
                pkgs_upgrade += groups_pkgs.get(group_name,[]).get('pkgs_upgrade',[]) 
            
            return pkgs_install,pkgs_upgrade
        except Exception as e:
            logging.error(str(e))
            return [],[]
    
    #计算升级或者安装这个包 是否存在需要卸载某个包
    def _make_delete_pkgs(self,cache,all_delete_pkgs):
        wouldDelete = cache._depcache.del_count
        deleted_pkgs = []
        if wouldDelete > len(all_delete_pkgs):
            for pkg_obj in cache:
                if pkg_obj.marked_delete:
                    if pkg_obj.name in all_delete_pkgs:
                        continue
                    else:
                        deleted_pkgs.append(pkg_obj.name)
            if deleted_pkgs:
                all_delete_pkgs += deleted_pkgs
        return deleted_pkgs

    #将获取本次升级的包 进行计算依赖关系 解决依赖问题
    def _make_problem_resolver(self,cache,pkgs_install = [],pkgs_upgrade = []):
        _success = True
        header = ''
        desc = ''

        #记录出现异常的包
        exception_pkg = ''

        last_broken_count = 0
        #计算出来的需要删除的包列表
        delete_pkgs = []
        delete_desc = []
        fresh_cache = Cache()
        try:
            logging.info("ProblemResolver install:%d , upgrade:%d",len(pkgs_install),len(pkgs_upgrade))
            logging.info("Start calculating dependencies...")
            
            #actiongroup 可以加速计算依赖关系 计算花费的时间将大幅度缩减
            with cache.actiongroup():
                if cache.get_changes():
                    cache.clear()
                resolver = apt.cache.ProblemResolver(cache)

                #标记计算所有需要安装的
                for pkg in pkgs_upgrade + pkgs_install:
                    exception_pkg = pkg

                    pkg_cache = cache[pkg]
                    #将第二个参数调整为False 当为True时就不能检查缺少依赖的包 默认自动移除掉了 
                    pkg_cache.mark_install(False, False, True)

                    #将可自动升级的标记为自动安装
                    if pkg_cache.is_upgradable == True:
                        auto = pkg_cache.is_auto_installed
                        pkg_cache.mark_auto(auto)

                    resolver.clear(pkg_cache)
                    resolver.protect(pkg_cache)
                    #出现破损的原因有 升级一个包需要安装其他的包或者需要删除其他的软件包
                    if cache._depcache.broken_count != last_broken_count and \
                            self.window_main.configs.getWithDefault("ConfigPkgStatus", "check_resover_remove", False) == True:
                        last_broken_count = cache._depcache.broken_count
                        tmp_pkgs,tmp_desc = self._emulate_calcul_delete(pkg,fresh_cache)
                        delete_pkgs.extend(tmp_pkgs)
                        delete_desc.extend(tmp_desc)
                
                resolver.resolve()
            return _success,delete_pkgs,delete_desc,header,desc
        except Exception as e:
            _success = False
            header = _("Could not calculate the upgrade")
            logging.error(header + "(" + exception_pkg + ")" +str(e))
            msg = get_broken_details(cache,False)
            pkg_string = ''
            for pkg in pkgs_install + pkgs_upgrade:
                pkg_string = pkg_string + ' ' + str(pkg)
            logging.info('Resolver calculation Packages List: '+pkg_string+'\n')
            desc ='\n' + msg
            logging.error('\n' + msg)
            threading_emulate = threading.Thread(target=emulate_calcul_broken,args=(pkgs_install + pkgs_upgrade,))
            threading_emulate.start()
            return _success,[],[],header,desc

    #调用aptdeamon结束之后处理的地方 不管是出错还是正常都在此处理
    def _action_done(self, action, is_cancelled,success, error_string,error_desc):
        #后端的状态 到空闲状态
        self.window_main.is_working = self.ACTION_DEFUALT_STATUS
        
        #升级完成后走的分支
        if action == self.ACTION_INSTALL:
            #记录这个过程中是否关机
            self.window_main.safe_manager.reset_safe()
            if self.now_upgrade.upgrade_mode != self.MODE_UPGRADE_SINGLE:
                self.window_main.configs.setValue("SystemStatus","isabnormalreboot",str(False),True)
            #解锁 禁止关机
            unLockedEnableShutdown()

            #单包安装模式 单独处理 目前 更新配置包和升级本身使用
            if self.now_upgrade.upgrade_mode == self.MODE_UPGRADE_SINGLE:
                if success:
                    #当单包升级的时候 升级本身时，让程序退出，再重新启动
                    if self.window_main.SELF_PKG_NAME in self.cache and self.cache[self.window_main.SELF_PKG_NAME] in self.cache.get_changes():
                        error_string = _("Update Manager upgrade is complete, please restart the control panel before performing the system update")
                        self.window_main.dbusController.DistUpdateDetectFinished(False,[self.window_main.SELF_PKG_NAME],error_string,error_desc)
                        #升级本身完成后 退出 有systemd 来进行重启服务
                        self.window_main.dbusController.Quit(None)
                    else:
                        #只有安装配置文件包 才会走到此处
                        self.window_main.start_available()
                    return
                else:
                    #当更新过程中安装配置包时 失败时发送错误信号
                    # 防止错误结果太长导致出现控制面板显示太多 目前最多20字符
                    if len(error_string + error_desc) > 20:
                        self.window_main.dbusController.DistUpdateDetectFinished(success,[],error_string,error_desc)
                    else:
                        self.window_main.dbusController.DistUpdateDetectFinished(success,[],error_string+' '+error_desc,error_desc)
                    return
            else:                
                upgrade_content = self.now_upgrade.upgrade_groups + self.now_upgrade.single_pkgs
                
                if self.is_version_upgrade == True and self.is_need_retry == True and success == False:
                    #增加重试次数的限制
                    if self.window_main.retry_limit != 0:
                        logging.warning("Retry the upgrade and installaton(%s) and retry number: %d",upgrade_content,self.window_main.retry_limit)
                        self.window_main.start_install(self.now_upgrade.upgrade_mode,True,upgrade_content)
                        self.window_main.retry_limit = self.window_main.retry_limit - 1
                        return

                false_num = self._make_insert_info(success,is_cancelled,self.now_upgrade,error_string,error_desc)
                
                if success == False:
                    logging.info("The number of failed installations or upgrades is %d",false_num)
                
                #修复bug 所有的包都安装成功了，但是却返回的失败，计算这种状况 从失败切换到安装成功的状态 
                #这个 为True的很低 大多数为False 只有当系统中apt 出现问题时 才会复现这种问题
                if success == False and false_num == 0 and upgrade_content != []:
                    logging.warning("Special Case switch success status from False to True...")
                    success = True

                if success:
                    #当组列表为空时 表示现在的单独进行安装某些包或卸载，不发信号到控制面板
                    #升级完成后从升级列表删除
                    for groups in self.now_upgrade.upgrade_groups:
                        self.upgrade_data.upgrade_groups.remove(groups)

                    for pkg in self.now_upgrade.single_pkgs:
                        self.upgrade_data.single_pkgs.remove(pkg)

                    error_string = ''
                    error_desc = ''

                #完成后清空cache
                if self.cache != None and self.cache.get_changes():
                    self.cache.clear()

                self.window_main.dbusController.UpdateInstallFinished(success,upgrade_content,error_string,error_desc)
                
                if success:
                    self.window_main.dbusController.CheckRebootRequired("self")
                    #安装完成之后 更新一次cache
                    logging.info("Install or Upgrade successful, so Now update Cache...")
                    self.window_main.only_update_cache()

        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_ALL:
            if success == False and 'The following signatures' in self.trans_error_details:
                error_desc = _("Check your source public key signature")
            #网络认证错误
            elif success == False and 'does the network require authentication?' in self.trans_error_details:
                error_desc = _("Check if your network requires authentication?")
            
            if success:
                #开始生成列表
                self.window_main.start_available()
            else:
                self.window_main.dbusController.DistUpdateDetectFinished(success,[],error_string+' '+error_desc,error_desc)

            if success == False:
                threading_emulate = threading.Thread(target=self.emulate_update)
                threading_emulate.start()

        elif action == self.ACTION_UPDATE and self.action_mode == self.MODE_UPDATE_CACHE:
                self.window_main.dbusController.UpdateDetectFinished(success,[''],error_string,error_desc)

        elif action == self.ACTION_FIX_BROKEN:
            self.window_main.dbusController.FixBrokenStatusChanged(True,success,100,'',error_string,error_desc)
            logging.warning("fix broken packages is complete...")
            
            if success:
                logging.info("Fix broken packages is complete to success...")
                self.window_main.start_available()
            else:
                self.window_main.dbusController.DistUpdateDetectFinished(False,[''],_("Check for update exceptions,fix system APT environment error."),error_string+' '+error_desc)
                logging.warning("fix broken packages is complete to failed...")
            # self.window_main.start_available()
        elif action == self.ACTION_FIX_INCOMPLETE:
            if success:
                logging.info("fix incomplete install success.")
            else:
                logging.warning("fix incomplete install failed.")

        elif action == self.ACTION_REMOVE_PACKAGES:
            self.window_main.dbusController.PurgePackagesFinished(success,error_string,error_desc)
        
        elif action == self.ACTION_INSTALL_DEB:
            import json
            #FIXME: '\r\n: \r\n\r\n'就认为是验证失败
            if success == False and '\r\n: \r\n\r\n' in self.trans_error_details:
                error_string = _("Package validation failed and installation was rejected.")
                error_desc = ''
            cache = Cache()
            self.window_main.dbusController.InstalldebFinished(success,error_string,error_desc)
            #软件商店/安装器上传
            if str(self.window_main.sqlite3_server.deb_metadata.get('source')) == 'kylin-software-center' \
                or str(self.window_main.sqlite3_server.deb_metadata.get('source')) == 'kylin-installer':
                status = 'success' if (success==True) else 'failed'
                UploadInfos  = {}
                UploadInfos.update({"packageName":str(self.window_main.sqlite3_server.deb_metadata.get('debname'))})
                UploadInfos.update({"source":str(self.window_main.sqlite3_server.deb_metadata.get('source'))})
                UploadInfos.update({"status":status})
                UploadInfos.update({"errorCode":str(error_string+" "+error_desc)})
                json_file = json.dumps(UploadInfos.copy())
                UploadInfos.pop("source")
                self.window_main.collector.UpdateMsg("UpdateInfos", json_file)

                UploadInfos.update({"old_version":str(self.window_main.sqlite3_server.deb_metadata.get('old_version'))})
                UploadInfos.update({"new_version":str(self.window_main.sqlite3_server.deb_metadata.get('new_version'))})
                json_file = json.dumps(UploadInfos.copy())
                self.window_main.collector.UpdateMsg("InstallInfos", json_file) 

                self.window_main.sqlite3_server.init_metadata()
                UploadInfos  = {}
            
        elif action == self.ACTION_BACKGROUND_UPGRADE:
            import json

            status = 'success' if (success==True) else 'failed'
            errorCode = str(error_string+" "+error_desc)
            UploadInfos  = {}
            upgrade_content = []
            cache = Cache()
            UploadInfos.update({"source":str(self.window_main.sqlite3_server.deb_metadata.get('source'))})
            UploadInfos.update({"status":status})
            UploadInfos.update({"errorCode":errorCode})

            if "background_list" in self.window_main.sqlite3_server.deb_metadata.keys():
                upgrade_content = self.window_main.sqlite3_server.deb_metadata["background_list"]
            
            self.window_main.dbusController.UpdateInstallFinished(success,upgrade_content,error_string,error_desc)

            if not self.window_main.sqlite3_server.check_background_upload(upgrade_content, status, errorCode):
                logging.info("no need upload background upgrade log.")
                return

            for bl in upgrade_content:
                UploadInfos.update({"packageName":str(bl)})
                
                current_version = ''
                if bl in cache:
                    pkg = cache[bl]
                    if pkg.installed:
                        current_version = pkg.installed.version
                UploadInfos.update({"new_version":str(current_version)})

                if 'background_list_old_versions' in self.window_main.sqlite3_server.deb_metadata.keys() and \
                            bl in self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].keys():
                    tmp = str(self.window_main.sqlite3_server.deb_metadata['background_list_old_versions'].get((bl)))
                    UploadInfos.update({"old_version":str(tmp)})
                else:
                    UploadInfos.update({"old_version":str("unknown")})
                json_file = json.dumps(UploadInfos.copy())
                self.window_main.collector.UpdateMsg("Background", json_file)

    #将安装完成的插入数据库 安装失败的计算那些包安装失败了 分类插入数据库中
    #将安装失败的数据进行返回
    def _make_insert_info(self,success,is_cancelled,_now_upgrade,error_string,error_desc):
        false_num = 0
        try:
            #目前去掉打印 错误日志
            # error_desc = error_desc +'\n\n'+self.trans_error_details
            #在安装失败 and 不是取消升级 则走下面的 否则不计算全部直接插入数据库
            if success == False and is_cancelled == False and self.now_upgrade.upgrade_mode != self.MODE_UPGRADE_SYSTEM:
                #获取当前系统中最新的cache
                fresh_cache = Cache(rootdir=self.window_main.cache.rootdir)
                #获取调整包列表 去掉版本号
                adjust_pkgs = [i.split("=")[0] for i in self.upgrade_data.adjust_pkgs]

                #升级存在单包更新 失败后计算每一个包是否安装完成 相应判断进行插入
                for pkg in _now_upgrade.single_pkgs:
                    pkg_obj = fresh_cache[pkg]
                    
                    #判断是否是源过滤调整的包 调整的话 判断安装版本 来解决是否安装成功
                    if pkg in adjust_pkgs:
                        if pkg_obj.is_installed == True and pkg_obj.is_now_broken == False and  pkg_obj.name + '=' + pkg_obj.installed.source_version in self.upgrade_data.adjust_pkgs:
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,True,'', '')
                        else:
                            false_num += 1
                            logging.warning("This single-package(%s) is not install or upgrade",pkg)
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,False,error_string, error_desc)
                    else:
                    #非调整版本的计算方式
                        if pkg_obj.is_installed == True and pkg_obj.is_upgradable == False and pkg_obj.is_now_broken == False:
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,True,'', '')
                        else:
                            false_num += 1
                            logging.warning("This single-package(%s) is not install or upgrade",pkg)
                            self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[pkg],[],self.upgrade_data.adjust_pkgs,False,error_string, error_desc)
                
                #组的计算是否升级成功的方式 将组按照单包来计算的
                if _now_upgrade.upgrade_groups != []:
                    pkgs_install,pkgs_upgrade = self._make_pkgs_list(self.cache,self.upgrade_data.upgrade_groups_pkgs,_now_upgrade.upgrade_groups,[])
                    group_error_log = ''
                    install_error_pkgs = []
                    total_pkg = pkgs_install + pkgs_upgrade

                    for pkg in pkgs_install + pkgs_upgrade:
                        pkg_obj = fresh_cache[pkg]
                        if pkg in adjust_pkgs:
                            if pkg_obj.is_installed == True and pkg_obj.is_now_broken == False and pkg_obj.installed.source_version + pkg_obj.name in self.upgrade_data.adjust_pkgs:
                                total_pkg.remove(pkg)
                            else:
                                install_error_pkgs.append(pkg)
                        else:
                            if pkg_obj.is_installed == True and pkg_obj.is_upgradable == False and pkg_obj.is_now_broken == False :
                                total_pkg.remove(pkg)
                            else:
                                install_error_pkgs.append(pkg)
                    
                    group_error_log = '\n'+ "This groups-package"+str(install_error_pkgs)+" is not install or upgrade" + '\n'
                    logging.warning(group_error_log)
                    #如果没有
                    if total_pkg == []:
                        self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[],_now_upgrade.upgrade_groups,self.upgrade_data.adjust_pkgs,True,'', '') 
                    else:
                        false_num += 1
                        self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,[],_now_upgrade.upgrade_groups,self.upgrade_data.adjust_pkgs,False,error_string,error_desc)
                #关闭临时cache
                fresh_cache.close()
            #其他情况直接写入数据库 不进行计算
            else:
                #取消的话全部是失败
                if is_cancelled == True:
                    false_num = len(_now_upgrade.single_pkgs) + len(_now_upgrade.upgrade_groups)
                self.window_main.sqlite3_server.insert_info(_now_upgrade.upgrade_mode,_now_upgrade.single_pkgs,_now_upgrade.upgrade_groups,self.upgrade_data.adjust_pkgs,success,error_string, error_desc)
                
            return false_num
        except Exception as e:
            logging.error(str(e))

    def emulate_update(self):
        args = ["apt-get", "update"]

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        logging.warning(str(p.stdout))
        # return p.stdout

# try aptdaemon
if os.path.exists("/usr/sbin/aptd") \
   and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ:
    # check if the gtkwidgets are installed as well
    try:
        from .InstallBackendAptdaemon import InstallBackendAptdaemon
    except ImportError:
        logging.exception("importing aptdaemon")

def get_backend(*args, **kwargs):
    """Select and return a package manager backend."""
    # try aptdaemon
    if (os.path.exists("/usr/sbin/aptd")
            and "UPDATE_MANAGER_FORCE_BACKEND_SYNAPTIC" not in os.environ):
        # check if the gtkwidgets are installed as well
        try:
            return InstallBackendAptdaemon(*args, **kwargs)
        except NameError:
            logging.exception("using aptdaemon failed")

    # nothing found, raise
    raise Exception("No working backend found, please try installing "
                    "aptdaemon or synaptic")
